/*
 * Author:     Yung-Pin Cheng
 * Purdue CS
 * LOG
 *
 *  Date:      03/16/98 version 2.0
 *             We can not prune the edge immediately because
 *             other path may borrow the edge to prune its future
 *             edges. So, in this version we try to keep table of flags
 *             whenever a path ever enter. 
 *
 *  Date:      03/15/98 version 1.2
 *             Consider one sequence at one time is wrong.
 *             We need to consider all the sequences together
 *             So, in this version, we will store all the sequences
 *             into a table and check all the sequences concurrently
 *           
 *  Date:      03/10/98 version 1.1
 *
 *             While testing fc2/elev/B series, I found that 
 *             pruning those disordered edges could cause a live state
 *             become a trap state and these states should be removed, too.
 *  Date:      02/18/98 version 1.0
 * 
 *             given an .dot file and a sequence of events that should
 *             happen in that order, the class return a .dot file with 
 *             those events violating the order pruned
 *

 */


import java.io.*;
import java.util.* ;

public class purge {
  static final int MAXKIND = 10 ;
  static final int MAXSEQ  = 10 ;
  String[][] seqtable = new String[MAXKIND][MAXSEQ] ;
  int seqno[]  = new int[MAXKIND] ;
  int kindno = 0 ;
  int oldlen = 0;

  BitSet bittable ;
  String[] namearray ;
  int[] bitindex ;
  boolean[] onoff ;

  int debugno = 1 ;
  int stackno = 0 ;

  public int isThisInSeq(String name) {
    //  given a string, it compare with all the string in sequence table
    //  if not found, it returns -1
    //  if found and it matches the ptr of that kind 
    //  it return 0 
    //  else return 1 for a violation of sequence 
    int i,j ;
    for (i=0;i < namearray.length ;i ++ )
      if (name.compareTo(namearray[i]) == 0 ) 
	 return i ;
    return -1 ;
  }
//---------------------------------------------------------------
public int[] clonePtr(int[] ptr) {
  int i ;
  int[] newptr  = new int[MAXKIND] ;
  for (i=0;i<kindno ;i++)
    newptr[i] = ptr[i] ;
  return newptr ;
}

private ret  Return(Stack stack) {
  if (stack.empty()) return null ;
  ret r = (ret) stack.pop();
  stackno -- ;
  return r ;
}
private void  Enter(Stack stack, ret r) {
  stack.push(r);
  stackno ++ ;
  if (stackno % 10  == 0) {
    System.out.print('\15');
    System.out.print("stack size usage: "+ stackno+ " items");
    System.out.flush();
  }
    
}

private void extendhistarray(node p, BitSet bitt) {
  int i, j ;
  BitSet[] tp ;
  // so the bitt contain a history that was not happening before
  j = p.histarray.length ;
  tp = p.histarray ;

  // free old histarray and create
  // a new histarry with one more element
  p.histarray = new BitSet[j+1] ; 
  for (i=0;i< tp.length ; i++ ) {
    p.histarray[i] = new BitSet(kindno);
    p.histarray[i].or(tp[i]);
  }
  p.histarray[i] = new BitSet(kindno);
  p.histarray[i].or(bitt);
}
  
  
  private void inithistarray(node p, BitSet bitt) {
    p.histarray = new BitSet[1] ;
    p.histarray[0] = new BitSet(kindno);
    p.histarray[0].or(bitt);
  }
  private boolean sameBittExisted(node p, BitSet bitt) {
    int j ;
    for (j=0;j< p.histarray.length ; j++)
      if (bitt.equals(p.histarray[j])) 
	return true ;
    return false ;
  }
	
  //---------------------------------------------------------------
public void depthFirstPurge(dotreader dr, 
			    node initial, 
			    BitSet bitt,  
			    String path) { 
  String err1 = new String("\nWARNING: A path come back to initial states\n"+
			  "However, the legal actions waited are different " +
			  "from the initialization. How could this happen ?"+
			  "I don't know");
  String warning1 = new String("\nDEBUG:We do have a deleted edge passing case");
  node p, np, oldp ;
  BitSet nbitt ;
  int code = -1 ;
  int i=0,j ;
  Stack stack  = new Stack();
  final int BEGINNING = 1 ;
  final int FORLOOP = 2 ;
  int retpoint ;
  ret r ;
  boolean err1out = false ;
  boolean warnout = false ;
  
  // Enter here, the depthfirst rountine contain a bitset table.
  // In commandline, how many pairs of edges is given, how many bits 
  // the bittable is.
  // For example, @ a0 a1 @ b0 b1 @ c0 c1, there are three bits.
  // bittable is 000 means a legal path is waiting for a0,b0,c0
  p = initial ;
  retpoint = BEGINNING ;
  for (;;) {
    switch(retpoint) {
    case  BEGINNING:
      // System.out.println("<<"+p.name+">>");
      if (p == dr.initnode && p.visited == true) {
	// A path now come back to the initial state.
	if (bitt.equals(bittable) == false) {
	  if (!err1out) { System.out.println(err1); err1out = true ;}
	}
	r = Return(stack); if (r == null) return ;
	p = r.p ; i = r.eno ; bitt = r.bitset ; retpoint = r.retpoint ;
	// System.out.println("go back to 1: back to "+p.name);
	continue ;
      }
      if (p.visited) {  // this state has been visited before
	if (sameBittExisted(p, bitt)) {
	  // we have been to this node before with the same bitset
	  oldp = p ;
	  r = Return(stack); if (r == null) return ;
	  p = r.p ; i = r.eno ; bitt = r.bitset ; retpoint = r.retpoint ;
	  //System.out.println(oldp.name +" visited, go back to "+p.name+
	  //		     "with i="+ i+ " retpoint="+ retpoint);
	  continue ;
	}
	extendhistarray(p, bitt);
      } else {
	p.visited = true ;    
	inithistarray(p,bitt) ;
      }
      i = -1 ; // begin first 
    case FORLOOP :
      i++ ;
      if (i >= p.eno ) {
	oldp = p ;
	r = Return(stack); if (r == null) return ;
	p = r.p ; i = r.eno ; bitt = r.bitset ; retpoint = r.retpoint ;
	// System.out.println("traverse "+ oldp.name+ " over, back to "+p.name);
	continue ;
      }
      //System.out.println("---------------["+debugno+ "]"+
      //		 "checking " +p.name +"' "+ dr.getEdge(p.ename[i]));
      debugno ++ ;
      code = isThisInSeq(dr.getEdge(p.ename[i])); //search the namearry
      if (code == -1 ) {  // this is an edge not in the sequence table
	np = dr.getNode(p.to[i]) ; // get the end node of the edge
	nbitt = (BitSet) bitt.clone();

	r = new ret(); r.p = p ; r.eno = i; r.bitset = bitt ; 
	r.retpoint = FORLOOP ; Enter(stack,r);
	p = np ; i = 0 ; bitt = nbitt ; retpoint = BEGINNING ;
	continue ;
	// depthFirstPrune(np, nbitt, path+"<"+e.edge.name+">"+np.name) ;
      } else {           // it is in one of the sequences
	// Here, code is the index of namearray
	// bitindex[code] translate the index into the index
	// of bit.

	if (bitt.get(bitindex[code]) == onoff[code] ) {
	  // this edge follow the sequence, then we flip the
	  // bit and continue depth first search
	  if (p.edgedeleted[i])  {
	    if (!warnout) { System.out.println(warning1); warnout = true; }
	  }
	  np = dr.getNode(p.to[i]);
	  nbitt = (BitSet) bitt.clone();
	  if (onoff[code]) nbitt.clear(bitindex[code]);
	  //the bit is true, so we set it off
	  else nbitt.set(bitindex[code]); //the bit is false, so we turn it on

	  r = new ret(); r.p = p ; r.eno = i; r.bitset = bitt ; 
  	  r.retpoint = FORLOOP ; Enter(stack,r);
	  p = np ; i = 0 ; bitt = nbitt ; retpoint = BEGINNING ;
	  continue ;
	  // depthFirstPrune(np, nbitt, path+"<"+e.edge.name+">"+np.name) ;
	} else {             // we have a case that violate the sequence
	  p.edgedeleted[i] = true ; // delete the EdgeWrap from the node p ;
	  // System.out.println("!!!!!"+p.name+"->"+p.to[i] + "killed");
	  retpoint = FORLOOP ; 
	  continue ;
	}
      }
    } // end of switch
  } // end of for (;;)
}
 //---------------------------------------------------------------
public void pruneReverseOrder(dotreader dr) {
  boolean somethingdeleted ;
  BitSet bitt ;
  
  int rno = 1 ;
  System.out.println("\n\nStart depth first search examining states... \n");
  dr.clearAllFlag();
  // initialize all the pointer of each kind
  bitt = (BitSet) bittable.clone();
  
  
  depthFirstPurge(dr, dr.initnode, bitt, dr.initnode.name) ;
  somethingdeleted= pruneUnreachable(dr);
  
  // cleanTrapState(gr);
  if (somethingdeleted) {
  } 
}
//---------------------------------------------------
public void depthFirstVisit(dotreader dr,node p) {
  int i ;
  p.visited = true ;
  for (i=0;i< p.eno; i++) {
    if (!p.edgedeleted[i] ) {
      node np = dr.getNode(p.to[i]);
      if (np.visited == false ) { // we haven't visited the node yet
	depthFirstVisit(dr, np);
      }
    }
  }
  }      
  
 //---------------------------------------------------
  public boolean pruneUnreachable(dotreader dr) {
    Enumeration e ;
    node p ;

    dr.clearVisitFlag(); // clear the flags first
    boolean rv = false ;
    depthFirstVisit(dr, dr.initnode);
    e = dr.nodet.elements();
    while (e.hasMoreElements()) {
      p = (node)e.nextElement();
      if (!p.visited && !p.deleted) {
	p.deleted = true ;
	rv = true ;
      }
    }
    return rv ;
  }
  //---------------------------------------------------
  public void processSequence(String args[]) {
    int i,j ;
    for (i=0;i< MAXKIND ; i++) {
      seqno[i] = 0 ;
    }
    i = 2 ; // the first parameter is the file name
    while (i < args.length) {
      if (args[i].compareTo("@") == 0 ) {
	i++ ; 
	if (i < args.length) {
	  while (args[i].compareTo("@") !=  0  ) {
	    seqtable[kindno][seqno[kindno]] = new String(args[i]);
	    i++ ; 
	    seqno[kindno] ++ ;
	    if (seqno[kindno] > 2 ) {
	      System.out.println("this version current only accept a pair"
				 + " of sequences ");
	      System.exit(1);
	    }
	      
	    if (i >= args.length )
	      break ;
	  }
	  kindno ++ ;
	}
      } else {
	System.out.println("usage: java purge source destination"+
			   " @ seq1 seq2 @ seq11 seq12 @ .... ");
	System.exit(1);
      }
    }
    System.out.println("The sequence table:");
    System.out.println("----------------------------------------------");
    bittable = new BitSet(kindno);
    namearray = new String[kindno * 2 ] ;
    bitindex = new int[kindno * 2];
    onoff = new boolean[kindno * 2] ;
    int k = 0 ;
    for (i=0;i< kindno ;i++ ) {
      for (j=0;j< seqno[i]; j++ ) {
	System.out.print(" "+seqtable[i][j]);
	namearray[k] = new String(seqtable[i][j]);
	bitindex[k] = (int) k/2 ;
	if (k%2 == 0 ) onoff[k] = false ;
	else onoff[k] = true ;
	k++ ;
      }
      System.out.println();
    }
    System.out.println("----------------------------------------------");
  }
 //---------------------------------------------------
  public static void main (String args[])  {
    if (args.length < 2 ) {
      System.out.println("Usage: java infile(.dot) outfile(.dot) @ seq1 seq2 @ seq21 seq22..");
      System.exit(1);
    }
    System.out.println("output:"+ args[1]);
    purge p = new purge() ;
    if (args.length == 2) {
      // no parameter is given, so we just write out the file 
      System.out.println("-------------------------------------------------");
      System.out.println(" no purge sequence is given, output same file ");
      System.out.println("-------------------------------------------------");
      dotreader d = new dotreader(args[0]);
      d.load(); // loading dot graph from files
      d.toFile(args[1]);
      System.out.println("\n\nPurge Program Successful\n");
      
    } else {
      p.processSequence(args);
      dotreader d = new dotreader(args[0]);
      d.load(); // loading dot graph from files
      
      p.pruneReverseOrder(d);
      
      d.toFile(args[1]);
      System.out.println("\n\nPurge Program Successful\n");
    }
  }
  
}







